#Const WRPN = True
' A "big integer" (unsigned) math library for Visual Basic.
'
' Carry_Bit is set or cleared on the following operations:
'   ShiftLeft
'   ShiftRight
'   Add
'   Subtract
'   Divide
'
' Over_Flow is set or cleared on the following operations:
'   Add
'   Subtract
'   Multiply
'   Divide
'   Power
'
' Loss_Of_Precision is set or cleared on the following operations:
'   Assign()'s
'   ToInteger
'   ToLong
'   ToDouble
'   ToUInt16
'   ToUInt32
'   ToUInt64
'
Public Class BigInt
    Private n() As Byte
    Private Over_Flow As Boolean
    Private Loss_Of_Precision As Boolean
    Private Carry_Bit As Boolean
    Private BitSize As Integer
    Private LastByte As Integer

    ' Standard constructor
    Public Sub New(ByVal size As Integer)
        If size <= 0 Then
            Throw New Exception("Size argument must be greater than zero")
            Exit Sub
        End If
        LastByte = (size - 1) \ 8
        BitSize = size
        ReDim n(LastByte)
    End Sub

    ' Integer constructor
    Public Sub New(ByVal size As Integer, ByVal int As Integer)
        Dim i As Integer
        Dim x As BigInt

        LastByte = (size - 1) \ 8
        ReDim n(LastByte)
        BitSize = size

        x = Assign(int, BitSize)
        For i = 0 To LastByte
            n(i) = x.n(i)
        Next
        Loss_Of_Precision = x.Loss_Of_Precision
    End Sub

#If Not WRPN Then
    ' Long constructor
    Public Sub New(ByVal size As Integer, ByVal lng As Long)
        Dim i As Integer
        Dim x As BigInt

        LastByte = (size - 1) \ 8
        ReDim n(LastByte)
        BitSize = size

        x = Assign(lng, BitSize)
        For i = 0 To LastByte
            n(i) = x.n(i)
        Next
        Loss_Of_Precision = x.Loss_Of_Precision
    End Sub

    ' Double constructor
    Public Sub New(ByVal size As Integer, ByVal dbl As Double)
        Dim i As Integer
        Dim x As BigInt

        LastByte = (size - 1) \ 8
        ReDim n(LastByte)
        BitSize = size

        x = Assign(dbl, BitSize)
        For i = 0 To LastByte
            n(i) = x.n(i)
        Next
        Loss_Of_Precision = x.Loss_Of_Precision
    End Sub

    ' UInt16 constructor
    Public Sub New(ByVal size As Integer, ByVal u16 As UInt16)
        Dim i As Integer
        Dim x As BigInt

        LastByte = (size - 1) \ 8
        ReDim n(LastByte)
        BitSize = size

        x = Assign(u16, BitSize)
        For i = 0 To LastByte
            n(i) = x.n(i)
        Next
        Loss_Of_Precision = x.Loss_Of_Precision
    End Sub

    ' UInt32 constructor
    Public Sub New(ByVal size As Integer, ByVal u32 As UInt32)
        Dim i As Integer
        Dim x As BigInt

        LastByte = (size - 1) \ 8
        ReDim n(LastByte)
        BitSize = size

        x = Assign(u32, BitSize)
        For i = 0 To LastByte
            n(i) = x.n(i)
        Next
        Loss_Of_Precision = x.Loss_Of_Precision
    End Sub

    ' UInt64 constructor
    Public Sub New(ByVal size As Integer, ByVal u64 As UInt64)
        Dim i As Integer
        Dim x As BigInt

        LastByte = (size - 1) \ 8
        ReDim n(LastByte)
        BitSize = size

        x = Assign(u64, BitSize)
        For i = 0 To LastByte
            n(i) = x.n(i)
        Next
        Loss_Of_Precision = x.Loss_Of_Precision
    End Sub
#End If

    ' *
    ' * Assignment from other "standard" variable types
    ' * 

    ' From integer
    Shared Function Assign(ByVal int As Integer, Optional ByVal size As Integer = 0) As BigInt
        Dim ans As BigInt
        Dim bytes() As Byte
        Dim i, last As Integer

        ans = New BigInt(IIf(size > 0, size, 32))

        bytes = BitConverter.GetBytes(int)
        last = IIf(ans.LastByte > UBound(bytes), UBound(bytes), ans.LastByte)

        For i = 0 To last
            ans.n(i) = bytes(i)
        Next

        ' Do we have any leftover bytes?
        ans.Loss_Of_Precision = False
        If last < UBound(bytes) Then
            For i = last + 1 To UBound(bytes)
                If bytes(i) <> 0 Then
                    ans.Loss_Of_Precision = True
                    Exit For
                End If
            Next
        End If
        ' How about any leftover bits?
        If ans.MaxBit() > ans.BitSize Then
            ans.Loss_Of_Precision = True
        End If

        ' Fill in the missing sign extension
        If UBound(bytes) < ans.LastByte And int < 0 Then
            For i = UBound(bytes) + 1 To ans.LastByte
                ans.n(i) = 255
            Next
        End If

        ans.Mask()
        Return ans
    End Function

    ' From long
    Shared Function Assign(ByVal lng As Long, Optional ByVal size As Integer = 0) As BigInt
        Dim ans As BigInt
        Dim bytes() As Byte
        Dim i, last As Integer

        ans = New BigInt(IIf(size > 0, size, 64))

        bytes = BitConverter.GetBytes(lng)
        last = IIf(ans.LastByte > UBound(bytes), UBound(bytes), ans.LastByte)

        For i = 0 To last
            ans.n(i) = bytes(i)
        Next

        ' Do we have any leftover bytes?
        ans.Loss_Of_Precision = False
        If last < UBound(bytes) Then
            For i = last + 1 To UBound(bytes)
                If bytes(i) <> 0 Then
                    ans.Loss_Of_Precision = True
                    Exit For
                End If
            Next
        End If
        ' How about any leftover bits?
        If ans.MaxBit() > ans.BitSize Then
            ans.Loss_Of_Precision = True
        End If

        ' Fill in the missing sign extension
        If UBound(bytes) < ans.LastByte And lng < 0 Then
            For i = UBound(bytes) + 1 To ans.LastByte
                ans.n(i) = 255
            Next
        End If

        ans.Mask()
        Return ans
    End Function

    ' From double (actually from long)
    Shared Function Assign(ByVal dbl As Double, Optional ByVal size As Integer = 0) As BigInt
        Dim ans As BigInt
        Dim temp As Long

        ans = New BigInt(IIf(size > 0, size, 64))

        ' Convert only if there is some useful value
        If dbl >= Long.MaxValue Or dbl <= Long.MinValue Then
            ' Humm... should this generate an exception?
            ans.Over_Flow = True
            Return ans
        End If
        ' Use default rounding rules!
        temp = CLng(dbl)

        ans = Assign(temp, size)
        Return ans
    End Function

#If Not WRPN Then
    ' From UInt16
    Shared Function Assign(ByVal u16 As UInt16, Optional ByVal size As Integer = 0) As BigInt
        Dim ans As BigInt
        Dim bytes() As Byte
        Dim i, last As Integer

        ans = New BigInt(IIf(size > 0, size, 16))

        bytes = BitConverter.GetBytes(u16)
        last = IIf(ans.LastByte > UBound(bytes), UBound(bytes), ans.LastByte)

        For i = 0 To last
            ans.n(i) = bytes(i)
        Next

        ' Do we have any leftover bytes?
        ans.Loss_Of_Precision = False
        If last < UBound(bytes) Then
            For i = last + 1 To UBound(bytes)
                If bytes(i) <> 0 Then
                    ans.Loss_Of_Precision = True
                    Exit For
                End If
            Next
        End If
        ' How about any leftover bits?
        If ans.MaxBit() > ans.BitSize Then
            ans.Loss_Of_Precision = True
        End If

        ans.Mask()
        Return ans
    End Function

    ' From UInt32
    Shared Function Assign(ByVal u32 As UInt32, Optional ByVal size As Integer = 0) As BigInt
        Dim ans As BigInt
        Dim bytes() As Byte
        Dim i, last As Integer

        ans = New BigInt(IIf(size > 0, size, 32))

        bytes = BitConverter.GetBytes(u32)
        last = IIf(ans.LastByte > UBound(bytes), UBound(bytes), ans.LastByte)

        For i = 0 To last
            ans.n(i) = bytes(i)
        Next

        ' Do we have any leftover bytes?
        ans.Loss_Of_Precision = False
        If last < UBound(bytes) Then
            For i = last + 1 To UBound(bytes)
                If bytes(i) <> 0 Then
                    ans.Loss_Of_Precision = True
                    Exit For
                End If
            Next
        End If
        ' How about any leftover bits?
        If ans.MaxBit() > ans.BitSize Then
            ans.Loss_Of_Precision = True
        End If

        ans.Mask()
        Return ans
    End Function

    ' From UInt64
    Shared Function Assign(ByVal u64 As UInt64, Optional ByVal size As Integer = 0) As BigInt
        Dim ans As BigInt
        Dim bytes() As Byte
        Dim i, last As Integer

        ans = New BigInt(IIf(size > 0, size, 64))

        bytes = BitConverter.GetBytes(u64)
        last = IIf(ans.LastByte > UBound(bytes), UBound(bytes), ans.LastByte)

        For i = 0 To last
            ans.n(i) = bytes(i)
        Next

        ' Do we have any leftover bytes?
        ans.Loss_Of_Precision = False
        If last < UBound(bytes) Then
            For i = last + 1 To UBound(bytes)
                If bytes(i) <> 0 Then
                    ans.Loss_Of_Precision = True
                    Exit For
                End If
            Next
        End If
        ' How about any leftover bits?
        If ans.MaxBit() > ans.BitSize Then
            ans.Loss_Of_Precision = True
        End If

        ans.Mask()
        Return ans
    End Function

    ' From a byte array
    Shared Function Assign(ByVal byte_array As Byte(), Optional ByVal size As Integer = 0) As BigInt
        Dim ans As BigInt
        Dim i, last As Integer

        ans = New BigInt(IIf(size > 0, size, (UBound(byte_array) + 1)) * 8)

        last = IIf(ans.LastByte > UBound(byte_array), UBound(byte_array), ans.LastByte)

        For i = 0 To last
            ans.n(i) = byte_array(i)
        Next

        Return ans
    End Function
#End If

    ' From another BigInt (no LOS detection!)
    Shared Function Assign(ByVal arg_x As BigInt, Optional ByVal size As Integer = 0) As BigInt
        Dim ans As BigInt
        Dim i, last As Integer

        ans = New BigInt(IIf(size > 0, size, arg_x.BitSize))

        last = IIf(ans.LastByte > arg_x.LastByte, arg_x.LastByte, ans.LastByte)

        For i = 0 To last
            ans.n(i) = arg_x.n(i)
        Next

        ' Copy the status flags too
        ans.Carry_Bit = arg_x.Carry_Bit
        ans.Over_Flow = arg_x.Over_Flow
        ans.Loss_Of_Precision = arg_x.Loss_Of_Precision

        ans.Mask()
        Return ans
    End Function

    ' From a String (We are a little creative... since the standard
    ' notation doesn't support a "&B" option). Errors are ignored.
    Shared Function Assign(ByVal s As String, Optional ByVal size As Integer = 0) As BigInt
        Dim i, bits, temp As Integer
        Dim n_byte, n_bit, finish As Integer
        Dim c, number As String
        Dim ans, z As BigInt
        ans = Nothing

        If s.Substring(0, 1) = "&" And s.Length > 2 Then
            number = Trim(s.Substring(2).ToLower)
            Select Case s.Substring(1, 1).ToLower
                Case "o"    ' Octal notation
                    ' What is the "native" size?
                    finish = 0
                    For i = number.Length - 1 To 0 Step -1
                        ' Stop at first unknown character
                        c = number.Substring(i, 1)
                        If c.IndexOfAny("01234567") = -1 Then
                            finish = finish + 1
                            Exit For
                        End If
                    Next
                    temp = (number.Length - finish) * 3
                    ' Let's round up to the nearest byte
                    n_bit = (((temp - 1) \ 8) + 1) * 8

                    ans = New BigInt(IIf(size > 0, size, n_bit))
                    z = New BigInt(ans.BitSize)

                    If temp > ans.BitSize Then
                        finish = (temp - ans.BitSize) \ 3
                    End If

                    ' Process the string
                    For i = number.Length - 1 To finish Step -1
                        If bits > ans.BitSize Then
                            ans.Loss_Of_Precision = True
                            Exit For
                        End If
                        temp = CInt("&O" & number.Substring(i, 1))
                        If temp <> 0 Then
                            ' Sneaky trick to prevent multiplication/shifts
                            z.Clear()
                            z.PutAt(temp, bits)
                            ans = Add(ans, z)
                        End If
                        ' Number of bits to shift
                        bits = bits + 3
                    Next
                Case "b"    ' Binary (not a really a standard notation!)
                    ' What is the "native" size?
                    finish = 0
                    For i = number.Length - 1 To 0 Step -1
                        ' Stop at first unknown character
                        c = number.Substring(i, 1)
                        If c.IndexOfAny("01") = -1 Then
                            finish = finish + 1
                            Exit For
                        End If
                    Next
                    temp = number.Length - finish
                    ' Let's round up to the nearest byte
                    n_bit = (((temp - 1) \ 8) + 1) * 8

                    ans = New BigInt(IIf(size > 0, size, n_bit))
                    ' Have to check LOS before we round off
                    If temp > IIf(size > 0, size, temp) Then
                        ans.Loss_Of_Precision = True
                    End If

                    If temp > ans.BitSize Then
                        finish = temp - ans.BitSize
                    End If

                    ' Process the string
                    For i = number.Length - 1 To finish Step -1
                        ' Which byte and which bit?
                        n_byte = (number.Length - 1 - i) \ 8
                        n_bit = (number.Length - 1 - i) Mod 8
                        If number.Substring(i, 1) = "1" Then
                            ans.n(n_byte) = ans.n(n_byte) + (2 ^ n_bit)
                        End If
                    Next
                Case "h"    ' Hexadecimal notation
                    ' What is the "native" size?
                    finish = 0
                    For i = number.Length - 1 To 0 Step -1
                        ' Stop at first unknown character
                        c = number.Substring(i, 1)
                        If c.IndexOfAny("0123456789abcdef") = -1 Then
                            finish = finish + 1
                            Exit For
                        End If
                    Next
                    temp = (number.Length - finish) * 4
                    ' Let's round up to the nearest byte
                    n_bit = (((temp - 1) \ 8) + 1) * 8

                    ans = New BigInt(IIf(size > 0, size, n_bit))
                    z = New BigInt(ans.BitSize)

                    If temp > ans.BitSize Then
                        finish = (temp - ans.BitSize) \ 4
                    End If

                    ' Process the string
                    For i = number.Length - 1 To finish Step -1
                        If bits > ans.BitSize Then
                            ans.Loss_Of_Precision = True
                            Exit For
                        End If
                        temp = CInt("&H" & number.Substring(i, 1))
                        If temp <> 0 Then
                            ' Sneaky trick to prevent multiplication/shifts
                            z.Clear()
                            z.PutAt(temp, bits)
                            ans = Add(ans, z)
                        End If
                        ' Number of bits to shift
                        bits = bits + 4
                    Next
            End Select
        Else
            ' If no leading "&", it must be a decimal notation
            finish = 0
            For i = s.Length - 1 To 0 Step -1
                ' Stop at first unknown character (including a negative sign)
                ' That means NO rounding!
                c = s.Substring(i, 1)
                If c.IndexOfAny("0123456789") = -1 Then
                    finish = finish + 1
                    Exit For
                End If
            Next
            ' Since we're not really doing bit manipulation, this size
            ' estimate isn't that critical
            temp = (s.Length - finish) * 3.5
            ' Let's round it off to the nearest byte
            n_bit = (((temp - 1) \ 8) + 1) * 8
            ans = New BigInt(IIf(size > 0, size, n_bit))

            Dim mult As New BigInt(ans.BitSize)
            Dim ten As New BigInt(ans.BitSize)
            Dim one As New BigInt(ans.BitSize)

            ' This is normal stuff.  You multiply each digit times the value
            ' associated with that decimal place, then add 'em all up.
            mult.n(0) = 1
            ten.n(0) = 10
            For i = s.Length - 1 To finish Step -1
                temp = CInt(s.Substring(i, 1))
                If temp <> 0 Then
                    ' Can't use my sneaky PutAt function...
                    z = Assign(temp, ans.BitSize)
                    ans = Add(ans, Multiply(z, mult))
                End If
                ' Decimal place
                mult = Multiply(mult, ten)
                If IsZero(mult) Then
                    ans.Loss_Of_Precision = True
                    Exit For
                End If
            Next

            ' If it's negative assume 2's complement
            If s.Substring(0, 1) = "-" Then
                one.n(0) = 1
                ans = Add(LogicalNot(ans), one)
            End If
        End If

        ans.Mask()
        Return ans
    End Function

    ' *
    ' * Conversion to "standard" data types
    ' *

    ' To Integer
    Public Function ToInteger() As Integer
        Dim x As BigInt

        x = Assign(Me, 32)
        Loss_Of_Precision = Not IsEqual(Me, x)

        Return BitConverter.ToInt32(x.n, 0)
    End Function

    ' To Long
    Public Function ToLong() As Long
        Dim x As BigInt

        x = Assign(Me, 64)
        Loss_Of_Precision = Not IsEqual(Me, x)

        Return BitConverter.ToInt64(x.n, 0)
    End Function

    ' To Double (actually from long)
    Public Function ToDouble() As Double
        Return CDbl(Me.ToLong())
    End Function

#If Not WRPN Then
    ' To UInt16
    Public Function ToUInt16() As UInt16
        Dim x As BigInt

        x = Assign(Me, 16)
        Loss_Of_Precision = Not IsEqual(Me, x)

        Return BitConverter.ToUInt16(x.n, 0)
    End Function

    ' To UInt32
    Public Function ToUInt32() As UInt32
        Dim x As BigInt

        x = Assign(Me, 32)
        Loss_Of_Precision = Not IsEqual(Me, x)

        Return (BitConverter.ToUInt32(x.n, 0))
    End Function

    ' To UInt64
    Public Function ToUInt64() As UInt64
        Dim x As BigInt

        x = Assign(Me, 64)
        Loss_Of_Precision = Not IsEqual(Me, x)

        Return BitConverter.ToUInt64(x.n, 0)
    End Function

    ' to Byte Array
    Public Function ToByteArray() As Byte()
        Dim i As Integer
        Dim ans(LastByte) As Byte

        For i = 0 To LastByte
            ans(i) = n(i)
        Next

        Return ans
    End Function
#End If

    ' *
    ' * To String functions
    ' *

    ' Override the default ToString() method (from Object Class)
    Public Overrides Function ToString() As String
        Return "&H" & ToStringHex()
    End Function

    ' To Hexadecimal string
    Public Function ToStringHex() As String
        Dim i, j, ending As Integer
        Dim ans As String = ""

        ' How many digits are possible 
        ending = (BitSize - 1) \ 4

        ' Instead of using the normal method, we use the GetAt function to
        ' extract the digit's value directly from the n() array.
        For i = 0 To ending
            ' Sneaky trick to prevent multiplication and division
            j = GetAt(i * 4, 4)
            If j > 9 Then
                ans = Chr(Asc("a") + j - 10) & ans
            Else
                ans = CStr(j) & ans
            End If
        Next

        Return ans
    End Function

    ' To Binary string
    Public Function ToStringBin(Optional ByVal pad As Boolean = False) As String
        Dim i, j As Integer
        Dim ans As String = ""

        For i = LastByte To 0 Step -1
            For j = 7 To 0 Step -1
                If (2 ^ j) And n(i) Then
                    ans = ans & "1"
                Else
                    ans = ans & "0"
                End If
            Next
            If i <> 0 And pad = True Then
                ans = ans & " "
            End If
        Next
        ' Trim to current word length, optionally separating the 8-bit words
        ' with a space
        If pad Then
            ans = ans.Substring(ans.Length - (BitSize + LastByte), BitSize + LastByte)
        Else
            ans = ans.Substring(ans.Length - BitSize, BitSize)
        End If

        Return ans
    End Function

    ' To Octal String
    Public Function ToStringOct() As String
        Dim i, j, ending As Integer
        Dim ans As String = ""

        ' How many digits are possible
        ending = (BitSize - 1) \ 3

        ' Instead of using the normal method, we use the GetAt function to
        ' extract the digit's value directly from the n() array.
        For i = 0 To ending
            ' Sneaky trick to prevent multiplication and division
            j = GetAt(i * 3, 3)
            ans = CStr(j) & ans
        Next

        Return ans
    End Function

    ' To Decimal String. This is the only type that has to deal with
    ' the "sign mode"... all others types (hex, oct, and bin) are inherently
    ' unsigned
    Public Function ToStringDec(Optional ByVal sign As Integer = 2) As String
        Dim i, j, ending As Integer
        Dim ten As New BigInt(BitSize)
        Dim tempx As New BigInt(BitSize)
        Dim one As New BigInt(BitSize)
        Dim z As New BigInt(BitSize)
        Dim is_neg As Boolean
        Dim ans As String = ""

        tempx = Assign(Me)
        is_neg = False

        Select Case sign
            Case 0
                ' Do nothing
            Case 1
                If SignBit() Then
                    is_neg = True
                    tempx = LogicalNot(Me)
                End If
            Case 2
                If SignBit() Then
                    is_neg = True
                    one.n(0) = 1
                    tempx = Add(LogicalNot(Me), one)
                End If
            Case Else
                Throw New Exception("Valid Sign arguments are 0=unsigned, 1=1's complement, and 2=2's complement")
                Return ans
        End Select

        ' How many digits are possible 
        ending = (BitSize - 1) / 3.5

        ' This normal stuff...  To get each digit, divide the number by 10
        ' and then get the remainder.  Repeat.
        ten.n(0) = 10
        For i = 0 To ending
            ' Can't use my sneaky GetAt function... Ugh!
            z = Modulus(tempx, ten)
            j = z.n(0)
            ans = CStr(j) & ans
            tempx = Divide(tempx, ten)
        Next

        If is_neg Then
            ans = "-" & ans
        End If

        Return ans
    End Function

    ' *
    ' * Logical Operators
    ' *

    ' And function
    Shared Function LogicalAnd(ByVal arg_x As Object, ByVal arg_y As Object, Optional ByVal size As Integer = 0) As BigInt
        Dim i As Integer
        Dim x, y As BigInt
        Dim ans As New BigInt(8)

        ' First we have to convert the input from whatever variable type
        ' it is... into a BigInt
        x = ans.Import(arg_x)
        y = ans.Import(arg_y)
        ' Based upon the native sizes of the converted input, pick the size
        ' of the answer
        ans.SetSize(IIf(size > 0, size, IIf(x.BitSize >= y.BitSize, x.BitSize, y.BitSize)))
        ' Resize the x and y (since the size may have changed)
        x.SetSize(ans.BitSize)
        y.SetSize(ans.BitSize)

        For i = 0 To ans.LastByte
            ans.n(i) = x.n(i) And y.n(i)
        Next

        Return ans
    End Function

    ' Or function
    Shared Function LogicalOr(ByVal arg_x As Object, ByVal arg_y As Object, Optional ByVal size As Integer = 0) As BigInt
        Dim i As Integer
        Dim x, y As BigInt
        Dim ans As New BigInt(8)

        x = ans.Import(arg_x)
        y = ans.Import(arg_y)
        ans.SetSize(IIf(size > 0, size, IIf(x.BitSize >= y.BitSize, x.BitSize, y.BitSize)))
        x.SetSize(ans.BitSize)
        y.SetSize(ans.BitSize)

        For i = 0 To ans.LastByte
            ans.n(i) = x.n(i) Or y.n(i)
        Next

        Return ans
    End Function

    ' Not function
    Shared Function LogicalNot(ByVal arg_x As Object, Optional ByVal size As Integer = 0) As BigInt
        Dim i As Integer
        Dim x As BigInt
        Dim ans As New BigInt(8)

        x = ans.Import(arg_x)
        ans.SetSize(IIf(size > 0, size, x.BitSize))
        x.SetSize(ans.BitSize)

        For i = 0 To ans.LastByte
            ans.n(i) = 255 - x.n(i)
        Next

        ans.Mask()
        Return ans
    End Function

    ' Exclusive Or
    Shared Function LogicalXor(ByVal arg_x As Object, ByVal arg_y As Object, Optional ByVal size As Integer = 0) As BigInt
        Dim i As Integer
        Dim x, y As BigInt
        Dim ans As New BigInt(8)

        x = ans.Import(arg_x)
        y = ans.Import(arg_y)
        ans.SetSize(IIf(size > 0, size, IIf(x.BitSize >= y.BitSize, x.BitSize, y.BitSize)))
        x.SetSize(ans.BitSize)
        y.SetSize(ans.BitSize)

        For i = 0 To ans.LastByte
            ans.n(i) = x.n(i) Xor y.n(i)
        Next

        Return ans
    End Function

    ' Bit-wise Shift Left
    Shared Function ShiftLeft(ByVal arg_x As Object, Optional ByVal size As Integer = 0) As BigInt
        Dim i, temp, carry As Integer
        Dim x As BigInt
        Dim ans As New BigInt(8)

        x = ans.Import(arg_x)
        ans.SetSize(IIf(size > 0, size, x.BitSize))
        x.SetSize(ans.BitSize)

        ' Carry is based upon the left most bit
        ans.Carry_Bit = x.SignBit()

        ' Shift left is the same as "multiply by 2"
        For i = 0 To ans.LastByte
            temp = (x.n(i) * 2) + carry
            ans.n(i) = temp Mod 256
            carry = temp \ 256
        Next

        ans.Mask()
        Return ans
    End Function

    ' Bit-wise Shift Right
    Shared Function ShiftRight(ByVal arg_x As Object, Optional ByVal size As Integer = 0) As BigInt
        Dim i, carry As Integer
        Dim x As BigInt
        Dim ans As New BigInt(8)

        x = ans.Import(arg_x)
        ans.SetSize(IIf(size > 0, size, x.BitSize))
        x.SetSize(ans.BitSize)

        ' Carry is based upon the right most bit
        ans.Carry_Bit = x.n(0) And 1

        ' Shift right is the same as "divide by 2"
        For i = ans.LastByte To 0 Step -1
            If i < ans.LastByte Then
                If x.n(i + 1) And 1 Then
                    carry = 128
                Else
                    carry = 0
                End If
            End If
            ans.n(i) = (x.n(i) \ 2) Or carry
        Next

        Return ans
    End Function

    ' *
    ' * Equality operators
    ' *

    ' Is Zero?
    Shared Function IsZero(ByVal arg_x As Object) As Boolean
        Dim i As Integer
        Dim x As BigInt
        Dim ans As New BigInt(8)

        x = ans.Import(arg_x)

        For i = 0 To x.LastByte
            If x.n(i) <> 0 Then
                Return False
            End If
        Next
        Return True
    End Function

    ' Is Equal?
    Shared Function IsEqual(ByVal arg_x As Object, ByVal arg_y As Object) As Boolean
        Dim i As Integer
        Dim x, y As BigInt
        Dim ans As New BigInt(8)

        x = ans.Import(arg_x)
        y = ans.Import(arg_y)

        If x.BitSize > y.BitSize Then
            y.SetSize(x.BitSize)
        End If
        If y.BitSize > x.BitSize Then
            x.SetSize(y.BitSize)
        End If

        For i = 0 To x.LastByte
            If x.n(i) <> y.n(i) Then
                Return False
            End If
        Next

        Return True
    End Function

    ' Is Less Than?
    Shared Function IsLess(ByVal arg_x As Object, ByVal arg_y As Object) As Boolean
        Dim i As Integer
        Dim x, y As BigInt
        Dim ans As New BigInt(8)

        x = ans.Import(arg_x)
        y = ans.Import(arg_y)

        If x.BitSize > y.BitSize Then
            y.SetSize(x.BitSize)
        End If
        If y.BitSize > x.BitSize Then
            x.SetSize(y.BitSize)
        End If

        For i = x.LastByte To 0 Step -1
            If x.n(i) > y.n(i) Then
                Return False
            End If
            If x.n(i) < y.n(i) Then
                Return True
            End If
        Next
        ' Must be equal
        Return False
    End Function

    ' Is Greater Than?
    Shared Function IsGreater(ByVal arg_x As Object, ByVal arg_y As Object) As Boolean
        Dim i As Integer
        Dim x, y As BigInt
        Dim ans As New BigInt(8)

        x = ans.Import(arg_x)
        y = ans.Import(arg_y)

        If x.BitSize > y.BitSize Then
            y.SetSize(x.BitSize)
        End If
        If y.BitSize > x.BitSize Then
            x.SetSize(y.BitSize)
        End If

        For i = x.LastByte To 0 Step -1
            If x.n(i) < y.n(i) Then
                Return False
            End If
            If x.n(i) > y.n(i) Then
                Return True
            End If
        Next
        ' Must be equal
        Return False
    End Function

    ' *
    ' * Simple math operators
    ' *

    ' Addition
    Shared Function Add(ByVal arg_x As Object, ByVal arg_y As Object, Optional ByVal size As Integer = 0) As BigInt
        Dim i, carry, temp, real_bitsize As Integer
        Dim x, y As BigInt
        Dim ans As New BigInt(8)

        x = ans.Import(arg_x)
        y = ans.Import(arg_y)
        ' We need a bit more "wiggle room" than just the current bitsize,
        ' so round off to the nearest byte and add 1 byte
        real_bitsize = IIf(size > 0, size, IIf(x.BitSize >= y.BitSize, x.BitSize, y.BitSize))
        ans.SetSize((((real_bitsize - 1) \ 8) + 2) * 8)
        x.SetSize(ans.BitSize)
        y.SetSize(ans.BitSize)

        ' Normal stuff... add two bytes, keep the "carry" for the next byte
        For i = 0 To ans.LastByte
            temp = CInt(x.n(i)) + CInt(y.n(i)) + carry
            ans.n(i) = temp Mod 256
            carry = temp \ 256
        Next

        ' We need wiggle room to detect overflow
        If ans.MaxBit() + 1 > real_bitsize Then
            ans.Over_Flow = True
            ans.Carry_Bit = True
        End If

        ans.SetSize(real_bitsize)
        Return ans
    End Function

    ' Subtraction
    Shared Function Subtract(ByVal arg_x As Object, ByVal arg_y As Object, Optional ByVal size As Integer = 0) As BigInt
        Dim i, carry, temp, real_bitsize As Integer
        Dim x, y As BigInt
        Dim ans As New BigInt(8)

        x = ans.Import(arg_x)
        y = ans.Import(arg_y)
        real_bitsize = IIf(size > 0, size, IIf(x.BitSize >= y.BitSize, x.BitSize, y.BitSize))
        ans.SetSize((((real_bitsize - 1) \ 8) + 2) * 8)
        x.SetSize(ans.BitSize)
        y.SetSize(ans.BitSize)

        ' Normal stuff... subtract two bytes, generate a "borrow" if needed
        For i = 0 To ans.LastByte
            temp = CInt(x.n(i)) - CInt(y.n(i)) - carry
            ' Do we need a "borrow"?
            If temp < 0 Then
                temp = temp + 256
                carry = 1
            Else
                carry = 0
            End If
            ans.n(i) = temp
        Next
        ' We need wiggle room to detect overflow
        If ans.MaxBit() + 1 > real_bitsize Then
            ans.Over_Flow = True
            ans.Carry_Bit = True
        End If

        ans.SetSize(real_bitsize)
        Return ans
    End Function

    ' Multiplication
    Shared Function Multiply(ByVal arg_x As Object, ByVal arg_y As Object, Optional ByVal size As Integer = 0) As BigInt
        Dim i, j, temp, real_bitsize As Integer
        Dim x, y, z As BigInt
        Dim ans As New BigInt(8)

        x = ans.Import(arg_x)
        y = ans.Import(arg_y)
        ' We need a bit more "wiggle room" than just the current bitsize,
        ' so round off to the nearest byte and add 1 byte
        real_bitsize = IIf(size > 0, size, IIf(x.BitSize >= y.BitSize, x.BitSize, y.BitSize))
        ans.SetSize((((real_bitsize - 1) \ 8) + 2) * 8)
        x.SetSize(ans.BitSize)
        y.SetSize(ans.BitSize)
        z = New BigInt(ans.BitSize)

        ' Let's handle some special cases
        If IsZero(x) Or IsZero(y) Then
            Return ans
        End If

        ' This is just like you were taught is school...  multiply each digit
        ' in the first number with every digit in the second.  Keep track of
        ' what "column" to write down the intermediate values, then add the
        ' intermediate values together.

        For i = 0 To x.MaxByte()
            For j = 0 To y.MaxByte()
                temp = CInt(x.n(i)) * CInt(y.n(j))
                ' anything to do?
                If temp <> 0 Then
                    If (i + j) <= ans.LastByte Then
                        ' This is a sneaky trick to create intermediate values
                        ' without using shifts.  Normally you'd multiply the
                        ' temp value times the value for that "decimal place".
                        z.Clear()
                        z.PutAt(temp, (i + j) * 8)
                        ans = Add(ans, z)
                    Else
                        ans.Over_Flow = True
                    End If
                End If
            Next
        Next
        ' We need wiggle room to detect over_flow
        If ans.MaxBit() + 1 > real_bitsize Then
            ans.Over_Flow = True
        End If

        ans.SetSize(real_bitsize)
        Return ans
    End Function

    ' Division
    Shared Function Divide(ByVal arg_x As Object, ByVal arg_y As Object, Optional ByVal size As Integer = 0) As BigInt
        Dim i, nx, ny, nc, ans_digit, real_bitsize As Integer
        Dim chunk, guess, tempx As BigInt
        Dim x, y, z As BigInt
        Dim ans As New BigInt(8)

        x = ans.Import(arg_x)
        y = ans.Import(arg_y)
        ' We need a bit more "wiggle room" than just the current bitsize,
        ' so round off to the nearest byte and add 1 byte
        real_bitsize = IIf(size > 0, size, IIf(x.BitSize >= y.BitSize, x.BitSize, y.BitSize))
        ans.SetSize((((real_bitsize - 1) \ 8) + 2) * 8)
        x.SetSize(ans.BitSize)
        y.SetSize(ans.BitSize)
        guess = New BigInt(ans.BitSize)
        chunk = New BigInt(ans.BitSize)

        ' Sanity check
        If IsZero(y) Then
            ans.Over_Flow = True
            Return ans
        End If

        ' Special case
        If IsLess(x, y) Then
            If Not IsZero(x) Then
                ans.Carry_Bit = True
            End If
            Return ans
        End If

        ' This is almost like you were taught in school... For each digit in
        ' the answer, guess the number of times the divisor will go into a
        ' "chunk" of the dividend.  We know the size of the chunk will have
        ' the same number of digits as x or be one digit larger.  Subtract
        ' the intermediate value and repeat.

        tempx = Assign(x)
        ny = y.MaxByte()
        Do
            ' If no more divisors... we're done
            If IsLess(tempx, y) Then
                Exit Do
            End If

            nx = tempx.MaxByte()
            ' Build a "chunk" that has the same number of digits as y
            chunk.Clear()
            For i = 0 To ny
                chunk.n(i) = tempx.n(i + nx - ny)
            Next
            nc = ny

            ' If it is too small, build another chunk with one more digit
            If IsLess(chunk, y) Then
                For i = 0 To ny + 1
                    chunk.n(i) = tempx.n(i + nx - ny - 1)
                Next
                nc = ny + 1
            End If

            ' To keep from testing all 255 possibilities for each "answer
            ' digit", we use a simplified binary tree search algorithm.
            ans_digit = 128
            guess.Clear()
            For i = 6 To 0 Step -1
                guess.n(0) = ans_digit
                ' This is why we need more wiggle room... so we don't have 
                ' to deal with overflows during the guessing of the "answer
                ' digit"
                z = Multiply(y, guess)
                If IsGreater(z, chunk) Then
                    ans_digit = ans_digit - (2 ^ i)
                Else
                    ans_digit = ans_digit + (2 ^ i)
                End If
            Next
            guess.n(0) = ans_digit
            z = Multiply(y, guess)
            If IsGreater(z, chunk) Then
                ans_digit = ans_digit - 1
            End If

            ' Place answer digit in the proper "column"
            z.Clear()
            z.PutAt(CInt(ans_digit), (nx - nc) * 8)

            ' Add it to the answer.
            ans = Add(ans, z)

            ' Generate the intermediate value by shifting the "last guess"
            ' multiplier to its proper place (which undoes the effect of the
            ' shift that occurred when we developed the chunk).
            guess.Clear()
            guess.n(nx - nc) = ans_digit
            z = Multiply(y, guess)

            ' Now subtract the intermediate value and continue...
            tempx = Subtract(tempx, z)
        Loop While nx - nc > 0

        ans.Carry_Bit = Not IsZero(tempx)

        ans.SetSize(real_bitsize)
        Return ans
    End Function

#If Not WRPN Then
    ' Power (raise x to the power of y)
    Shared Function Power(ByVal arg_x As Object, ByVal arg_y As Object, Optional ByVal size As Integer = 0) As BigInt
        Dim real_bitsize As Integer
        Dim over As Boolean
        Dim x, y, one As BigInt
        Dim ans As New BigInt(8)

        x = ans.Import(arg_x)
        y = ans.Import(arg_y)
        ' We need a bit more "wiggle room" than just the current bitsize,
        ' so round off to the nearest byte and add 1 byte
        real_bitsize = IIf(size > 0, size, IIf(x.BitSize >= y.BitSize, x.BitSize, y.BitSize))
        ans.SetSize((((real_bitsize - 1) \ 8) + 2) * 8)
        x.SetSize(ans.BitSize)
        y.SetSize(ans.BitSize)

        one = Assign(1, ans.BitSize)

        ' Special cases
        If IsZero(y) Then
            ' Any number raised to the 0th power is 1
            ans.n(0) = 1
            Return ans
        End If
        If IsZero(x) Then
            Return ans
        End If

        ' Just mutliply x with itself that many times
        ans = Assign(x)
        y = Subtract(y, one)
        Do While Not IsZero(y)
            ans = Multiply(ans, x)
            If ans.Over_Flow Then
                over = True
            End If
            y = Subtract(y, one)
        Loop

        If over Or ans.MaxBit() - 1 > real_bitsize Then
            ans.Over_Flow = True
        End If

        ans.SetSize(real_bitsize)
        Return ans
    End Function
#End If

    ' *
    ' * Some not-so common math functions
    ' *

    ' Modulus (integer remainder after division)
    Shared Function Modulus(ByVal arg_x As Object, ByVal arg_y As Object, Optional ByVal size As Integer = 0) As BigInt
        Dim i, nx, ny, nc, ans_digit, real_bitsize As Integer
        Dim chunk, guess, tempx As BigInt
        Dim x, y, z As BigInt
        Dim ans As New BigInt(8)

        x = ans.Import(arg_x)
        y = ans.Import(arg_y)
        ' We need a bit more wiggle room than the just current bitsize.
        real_bitsize = IIf(size > 0, size, IIf(x.BitSize >= y.BitSize, x.BitSize, y.BitSize))
        ans.SetSize((((real_bitsize - 1) \ 8) + 2) * 8)
        x.SetSize(ans.BitSize)
        y.SetSize(ans.BitSize)
        guess = New BigInt(ans.BitSize)
        chunk = New BigInt(ans.BitSize)

        ' Sanity check
        If IsZero(y) Then
            ans.Over_Flow = True
            Return ans
        End If

        ' Special case (if x < y, then return x)
        If IsLess(x, y) Then
            Return x
        End If

        ' This is almost like you were taught in school... For each digit in
        ' the answer, guess the number of times the divisor will go into a
        ' "chunk" of the dividend.  We know the size of the chunk will have
        ' the same number of digits as x or be one digit larger.  Subtract
        ' the intermediate value and repeat.

        tempx = Assign(x)
        ny = y.MaxByte()
        Do
            ' If no more divisors... we're done
            If IsLess(tempx, y) Then
                Exit Do
            End If

            nx = tempx.MaxByte()
            ' Build a "chunk" that has the same number of digits as y
            chunk.Clear()
            For i = 0 To ny
                chunk.n(i) = tempx.n(i + nx - ny)
            Next
            nc = ny

            ' If it is too small, build another chunk with one more digit
            If IsLess(chunk, y) Then
                For i = 0 To ny + 1
                    chunk.n(i) = tempx.n(i + nx - ny - 1)
                Next
                nc = ny + 1
            End If

            ' To keep from testing all 255 possibilities for each "answer
            ' digit", we use a simplified binary tree search algorithm.
            ans_digit = 128
            guess.Clear()
            For i = 6 To 0 Step -1
                guess.n(0) = ans_digit
                ' This is why we need more wiggle room... so we don't have 
                ' to deal with overflows during the guessing of the "answer
                ' digit"
                z = Multiply(y, guess)
                If IsGreater(z, chunk) Then
                    ans_digit = ans_digit - (2 ^ i)
                Else
                    ans_digit = ans_digit + (2 ^ i)
                End If
            Next
            guess.n(0) = ans_digit
            z = Multiply(y, guess)
            If IsGreater(z, chunk) Then
                ans_digit = ans_digit - 1
            End If

            ' Generate the intermediate value by shifting the "last guess"
            ' multiplier to its proper place (which undoes the effect of the
            ' shift that occurred when we developed the chunk).
            guess.Clear()
            guess.n(nx - nc) = ans_digit
            z = Multiply(y, guess)

            ' Now subtract the intermediate value and continue...
            tempx = Subtract(tempx, z)
        Loop While nx - nc > 0

        tempx.SetSize(real_bitsize)
        Return tempx
    End Function

    ' Square Root using Newton's method
    Shared Function SquareRoot(ByVal arg_x As Object, Optional ByVal size As Integer = 0) As BigInt
        Dim x, r As BigInt
        Dim ans As New BigInt(8)

        x = ans.Import(arg_x)
        ans.SetSize(IIf(size > 0, size, x.BitSize))
        x.SetSize(ans.BitSize)
        r = New BigInt(ans.BitSize)

        If IsZero(x) Then
            ans.Over_Flow = True
            Return ans
        End If

        ans = Assign(x)
        r.n(0) = 1
        Do While IsGreater(ans, r)
            r = Divide(x, ans)
            ' Using right shift instead of division by 2
            ans = ShiftRight(Add(ans, r))
        Loop

        ' Carry on remainder (not a perfect square)
        r = Subtract(x, Multiply(ans, ans))
        ans.Carry_Bit = Not IsZero(r)

        Return ans
    End Function

    ' *
    ' * Some "utility" functions
    ' *

    ' Return the "status flags" of the variable
    Public Function Overflow() As Boolean
        Return Over_Flow
    End Function

    Public Function LossOfPrecision() As Boolean
        Return Loss_Of_Precision
    End Function

    Public Function CarryBit() As Boolean
        Return Carry_Bit
    End Function

    ' Get or Set the Size of the variable.  This does NOT trigger
    ' a Loss_Of_Precision
    Public Sub SetSize(ByVal size As Integer)
        LastByte = (size - 1) \ 8
        ReDim Preserve n(LastByte)
        BitSize = size
        Mask()
    End Sub

    Public Function GetSize() As Integer
        Return BitSize
    End Function

    ' Returns the location of the most significant bit
    Public Function MaxBit() As Integer
        Dim i, j As Integer

        For i = LastByte To 0 Step -1
            For j = 7 To 0 Step -1
                If n(i) And (2 ^ j) Then
                    Return (i * 8) + j
                End If
            Next
        Next
        ' Must be zero
        Return 0
    End Function

    ' *
    ' * Some private functions that are used throughout this class
    ' *

    ' Translate one of the know variable types into a BigInt
    Private Function Import(ByVal obj As Object) As BigInt
        Dim ans As BigInt
        Dim name As String

        name = obj.GetType().Name
        Select Case name
            Case "Int32"
                ans = Assign(CInt(obj))
            Case "Int64"
                ans = Assign(CLng(obj))
            Case "Double"
                ans = Assign(CDbl(obj))
            Case "String"
                ans = Assign(CStr(obj))
#If Not WRPN Then
            Case "UInt16"
                ans = Assign(CType(obj, System.UInt16))
            Case "UInt32"
                ans = Assign(CType(obj, System.UInt32))
            Case "UInt64"
                ans = Assign(CType(obj, System.UInt64))
#End If
            Case "BigInt"
                ans = Assign(CType(obj, BigInt))
            Case Else
                Throw New Exception("Conversion from type " & name & " is not supported")
        End Select

        Return ans
    End Function

    ' Make it Zero
    Private Sub Clear()
        Dim i As Integer

        For i = 0 To LastByte
            n(i) = 0
        Next

        ' Also reset the status flags
        Carry_Bit = False
        Over_Flow = False
        Loss_Of_Precision = False
    End Sub

    ' Returns the location of the most significant digit (byte) in
    ' the n() array
    Private Function MaxByte() As Integer
        Dim i As Integer

        For i = LastByte To 0 Step -1
            If n(i) <> 0 Then
                Return i
            End If
        Next

        Return 0
    End Function

    ' Put the integer value at the specified starting bit location within
    ' the n() array.  This prevents a lot of bit shifting or multiplication,
    ' when you can place the value where you want it.
    Private Sub PutAt(ByVal i As Integer, ByVal Location As Integer)
        Dim n_bit, n_byte As Integer
        Dim temp As Long

        n_byte = Location \ 8
        n_bit = Location Mod 8
        temp = i * (2 ^ n_bit)

        n(n_byte) = temp Mod 256
        If temp > 255 Then
            If n_byte < LastByte Then
                n(n_byte + 1) = temp \ 256
            Else
                Over_Flow = True
            End If
        End If
    End Sub

    ' Get an integer value from the specified location (and length) within
    ' the n() array.  Sorta like the Mid$ function for the n() array.
    Private Function GetAt(ByVal Location As Integer, ByVal Length As Integer) As Integer
        Dim ans, temp, n_bit, n_byte As Integer

        n_byte = Location \ 8
        n_bit = Location Mod 8

        If n_byte < LastByte Then
            temp = CInt(n(n_byte)) + (CInt(n(n_byte + 1)) * 256)
        Else
            temp = CInt(n(n_byte))
        End If
        ans = (temp \ (2 ^ n_bit)) Mod (2 ^ Length)

        Return ans
    End Function

    ' Mask the current value to the proper number of bits 
    Private Sub Mask()
        Dim mask_value As Integer

        ' Anything to do?
        If BitSize = (LastByte + 1) * 8 Then
            Exit Sub
        End If

        mask_value = ((2 ^ ((BitSize - 1) Mod 8)) * 2) - 1

        ' Mask that last byte
        n(LastByte) = n(LastByte) And mask_value
    End Sub

    ' Check to see if the Sign Bit is set
    Private Function SignBit() As Boolean
        Dim mask_value As Integer

        mask_value = 2 ^ ((BitSize - 1) Mod 8)

        If n(LastByte) And mask_value Then
            Return True
        End If

        Return False
    End Function

    ' *
    ' * Functions that are specific to the WRPN Project
    ' *
#If WRPN Then

    ' Split into two parts (even bit sizes only!). The size of the
    ' 2 halves (called by reference) is determined by the size
    ' of this instance
    Public Sub Split(ByRef Left_Hand As BigInt, ByRef Right_Hand As BigInt)
        Dim temp As BigInt
        Dim i, half_bitsize, half_maxbyte As Integer

        ' A wee bit of sanity checking
        If (BitSize Mod 2) Then
            Throw New Exception("Bits not divisible by two")
        End If
        half_bitsize = BitSize / 2
        half_maxbyte = (half_bitsize - 1) \ 8

        Right_Hand.SetSize(half_bitsize)
        Left_Hand.SetSize(half_bitsize)

        ' Right size (is easy)
        For i = 0 To half_maxbyte
            Right_Hand.n(i) = n(i)
        Next
        Right_Hand.Mask()

        ' Left side (right shift "half_bitsize" number of times)
        temp = Divide(Me, 2 ^ half_bitsize)
        For i = 0 To half_maxbyte
            Left_Hand.n(i) = temp.n(i)
        Next
        Left_Hand.Mask()
    End Sub

    ' Combine two halves into one value.  The size of the resulting
    ' value is determined by the size of calling arguments.  Modifies
    ' the instance directly.
    Public Sub Combine(ByVal Left_Hand As BigInt, ByVal Right_Hand As BigInt)
        Dim temp As BigInt
        Dim i As Integer

        ' A wee bit of sanity checking
        If (Left_Hand.BitSize <> Right_Hand.BitSize) Then
            Throw New Exception("Bits sizes do not match")
        End If

        Me.SetSize(Left_Hand.BitSize * 2)

        ' Right side
        For i = 0 To Right_Hand.MaxByte
            n(i) = Right_Hand.n(i)
        Next

        ' Left side (shift left "left_hand.bitsize" number of times)
        temp = Multiply(Left_Hand, 2 ^ Left_Hand.BitSize, BitSize)
        For i = 0 To temp.MaxByte
            n(i) = n(i) Or temp.n(i)
        Next
    End Sub

    ' Count the number of set bits.
    Public Function SumBits() As Integer
        Dim i, j, ans As Integer

        For i = LastByte To 0 Step -1
            For j = 7 To 0 Step -1
                If n(i) And (2 ^ j) Then
                    ans = ans + 1
                End If
            Next
        Next

        Return ans
    End Function

    ' Create a sign mask of the given size
    Shared Function MakeSignMask(ByVal mask_size As Integer, ByVal size As Integer) As BigInt
        Dim ans As New BigInt(size)
        Dim i, last, mask_value As Integer

        last = (mask_size - 1) \ 8
        mask_value = 2 ^ ((mask_size - 1) Mod 8)

        For i = 0 To last - 1
            ans.n(i) = 0
        Next
        ans.n(last) = mask_value

        Return ans
    End Function

    ' Create a word mask of the given size
    Shared Function MakeWordMask(ByVal mask_size As Integer, ByVal size As Integer) As BigInt
        Dim ans As New BigInt(size)
        Dim i, last, mask_value As Integer

        last = (mask_size - 1) \ 8
        ' word mask is (sign mask * 2) -1
        mask_value = ((2 ^ ((mask_size - 1) Mod 8)) * 2) - 1

        For i = 0 To last - 1
            ans.n(i) = 255
        Next
        ans.n(last) = mask_value

        Return ans
    End Function

    ' Public version of the private SignBit function
    Public Function IsNegative() As Boolean
        Return SignBit()
    End Function
#End If
End Class
